"""
Loader implementation for automatic compilation and linking of C-API extensions upon import.
"""
import sys
import os.path
import imp
import subprocess
from . import lib

suffixes = [
	('.py.c', 'rb', None),
	('.py.m', 'rb', None),
	('.py.c++', 'rb', None),
]

if hasattr(imp, 'cache_from_source'):
	def cache_path(path):
		return imp.cache_from_source(path)[:-len('.pyc')]
else:
	def cache_path(path):
		return path[:path.rfind('.py.')]

class CompilationError(ImportError):
	def __init__(self,
		module_path, source_path,
		compile_log, return_code,
	):
		self.module_path = module_path
		self.source_path = source_path
		self.compile_log = compile_log
		self.return_code = return_code

	def __str__(self):
		msg = "could not compile '{0.module_path}' from source '{0.source_path}'\n".format(self)
		msg += "*** compiler exited with code '{0.return_code}' and emitted the following:\n  ".format(self)
		msg += '  '.join(self.compile_log.splitlines(True))
		msg += '\n***'
		return msg

class LinkError(ImportError):
	def __init__(self,
		module_path, source_path,
		compile_log, return_code,
	):
		self.module_path = module_path
		self.source_path = source_path
		self.compile_log = compile_log
		self.return_code = return_code

	def __str__(self):
		msg = "could not link '{0.module_path}' from object '{0.source_path}'\n".format(self)
		msg += "*** linker exited with code '{0.return_code}' and emitted the following:\n  ".format(self)
		msg += '  '.join(self.compile_log.splitlines(True))
		msg += '\n***'
		return msg

class CLoader(object):
	"""
	Compile and Linke C-API modules with import statements.
	"""
	target = sys.platform

	stages = [
		'compile',
		'link',
	]

	types = [
		('c', '.py.c'),
		('objc', '.py.m'),
		('c++', '.py.c++'),
	]

	def logfile(self, stage):
		return '.'.join((self.cprefix, stage, 'log'))

	def execute(self, stage, command, popen = subprocess.Popen):
		logfile = self.logfile(stage)
		if os.path.exists(logfile):
			os.unlink(logfile)
		with open(logfile, 'a') as log:
			log.write('\n--------\n')
			log.write(' '.join(command))
			log.write('\n--------\n\n')
			log.flush()
			proc = popen(command, stdout = log, stderr = subprocess.STDOUT)
			# note command being ran
			return proc.wait()

	def __init__(self, pkg, name, source, type = 'c'):
		self.type = type
		self.package = pkg
		self.name = name
		self.fullname = '.'.join((pkg,name))
		self.source = source
		self.cprefix = cache_path(source) + sys.abiflags
		self.cof = self.cprefix + '.o'
		self.dll = self.cprefix + '.so'

	@classmethod
	def find_module(typ, fullname, paths = None,
		realpath = os.path.realpath,
		join = os.path.join,
		exists = os.path.exists,
	):
		if paths is None:
			# XXX: might as well support top-level...
			return None

		names = fullname.split('.')
		modname = names[-1]
		for f in paths:
			path = join(realpath(f), modname)
			for type, ext in typ.types:
				source = path + ext
				if exists(source):
					return typ('.'.join(names[:-1]), modname, source, type = type)

	def get_code(self, module):
		pass

	def get_data(self, module):
		pass

	def get_filename(self, name):
		return self.source

	def get_source(self, name):
		with open(self.source) as f:
			return f.read()

	def is_package(self, *args, **kw):
		'cloader does not handle packages'
		return False

	def build(self):
		dir = os.path.dirname(self.cprefix)
		if not os.path.exists(dir):
			os.mkdir(dir)

		rc = self.execute('compile', lib.compile(self.cof, self.source, self.type))
		if rc != 0 or not os.path.exists(self.cof):
			with open(self.logfile('compile'), 'rb') as log:
				output = log.read().decode('utf-8', errors='replace').strip()
			raise CompilationError(self.fullname, self.source, output, rc)

		rc = self.execute('link', lib.link(self.dll, self.cof, self.type))
		if rc != 0 or not os.path.exists(self.dll):
			with open(self.logfile('link'), 'rb') as log:
				output = log.read().decode('utf-8', errors='replace').strip()
			raise LinkError(self.fullname, self.source, output, rc)

	def load_module(self, fullname, exists = os.path.exists, getmtime = os.path.getmtime):
		# instance exists, so compilation succeeded
		if fullname in sys.modules:
			return sys.modules[fullname]
		if not exists(self.dll) or getmtime(self.dll) < getmtime(self.source):
			self.build()
		# compiled "successfully" or already compiled.
		mod = sys.modules[fullname] = imp.load_dynamic(self.name, self.dll)
		mod.__source__ = self.source
		mod.__cached__ = self.dll
		mod.__loader__ = self
		return mod

# recognize .py.c, .py.m, and .py.c++ files.
_original_get_suffixes = imp.get_suffixes

def get_suffixes(*args, _suffixes = suffixes, **kw):
	'c.loader override of imp.get_suffixes'
	sl = _original_get_suffixes()
	sl.extend(_suffixes)
	return sl

def install():
	'Install the meta path hook and override imp.get_suffixes'
	if CLoader not in sys.meta_path:
		sys.meta_path.append(CLoader)
		imp.get_suffixes = get_suffixes

def remove():
	'Remove the meta path hook.'
	sys.meta_path.remove(CLoader)
	imp.get_suffixes = _original_get_suffixes

#: Contents of the cloader.pth file.
pthfile = """# THIS FILE MAKES THE *.py.(c|m|c++) FILES IMPORTABLE
import c.loader; c.loader.install()
"""

def autoload(contents = pthfile, filename = 'c.pth'):
	import site
	dir = site.getsitepackages()[0]
	with open(os.path.join(dir, filename), 'w') as f:
		f.write(contents)

if __name__ == '__main__':
	autoload()
